@@ -29,6 +29,7 @@ import concurrent.futures
 from s3tethys import utils
 
 so_http.DEFAULT_BUFFER_SIZE = 524288
+AnyPath = Union[str, pathlib.Path]
 
 ###################################################
 ### Parameters
@@ -78,7 +79,7 @@ def s3_client(connection_config: dict, max_pool_connections: int = 30):
     return s3
 
 
-def get_object_s3(obj_key: str, bucket: str, s3: botocore.client.BaseClient = None, connection_config: dict = None, public_url: HttpUrl=None, version_id=None, range_start: int=None, range_end: int=None, chunk_size=524288):
+def get_object_s3(obj_key: str, bucket: str, s3: botocore.client.BaseClient = None, connection_config: dict = None, public_url: HttpUrl=None, version_id: str=None, range_start: int=None, range_end: int=None, chunk_size: int=524288, retries=3):
     """
     General function to get an object from an S3 bucket. One of s3, connection_config, or public_url must be used.
 
@@ -86,14 +87,22 @@ def get_object_s3(obj_key: str, bucket: str, s3: botocore.client.BaseClient = No
     ----------
     obj_key : str
         The object key in the S3 bucket.
+    bucket : str
+        The bucket name.
     s3 : botocore.client.BaseClient
         An S3 client object created via the s3_client function.
     connection_config : dict
         A dictionary of the connection info necessary to establish an S3 connection. It should contain service_name, s3, endpoint_url, aws_access_key_id, and aws_secret_access_key.
-    public_url : str
-        A URL to a public S3 bucket. This is generally only used for Backblaze object storage.
-    bucket : str
-        The bucket name.
+    public_url : http str
+        A URL to a public S3 bucket.
+    version_id : str
+        The S3 version id associated with the object.
+    range_start: int
+        The byte range start for the file.
+    range_end: int
+        The byte range end for the file.
+    chunk_size: int
+        The amount of bytes to download as once.
 
     Returns
     -------
@@ -126,10 +135,25 @@ def get_object_s3(obj_key: str, bucket: str, s3: botocore.client.BaseClient = No
     if isinstance(public_url, str) and (version_id is None):
         url = utils.create_public_s3_url(public_url, bucket, obj_key)
 
+        transport_params['timeout'] = 120
+
         if range1 is not None:
             transport_params['headers'] = {'Range': range1}
 
-        file_obj = smart_open.open(url, 'rb', transport_params=transport_params, compression='disable')
+        counter = retries
+        while True:
+            try:
+                file_obj = smart_open.open(url, 'rb', transport_params=transport_params, compression='disable')
+                break
+            except Exception as err:
+                counter = counter - 1
+                if counter > 0:
+                    sleep(3)
+                else:
+                    print('smart_open could not open url with the following error:')
+                    print(err)
+                    file_obj = None
+                    break
 
     elif isinstance(s3, botocore.client.BaseClient) or isinstance(connection_config, dict):
         if range1 is not None:
@@ -143,7 +167,12 @@ def get_object_s3(obj_key: str, bucket: str, s3: botocore.client.BaseClient = No
         s3_url = s3_url_base.format(bucket=bucket, key=obj_key)
         transport_params['client'] = s3
 
-        file_obj = smart_open.open(s3_url, 'rb', transport_params=transport_params, compression='disable')
+        try:
+            file_obj = smart_open.open(s3_url, 'rb', transport_params=transport_params, compression='disable')
+        except Exception as err:
+            print('smart_open could not open url with the following error:')
+            print(err)
+            file_obj = None
 
     else:
         raise TypeError('One of s3, connection_config, or public_url needs to be correctly defined.')
@@ -151,9 +180,81 @@ def get_object_s3(obj_key: str, bucket: str, s3: botocore.client.BaseClient = No
     return file_obj
 
 
-def stream_to_file(file_obj, file_path, chunk_size=524288):
+def url_to_stream(url: HttpUrl, range_start: int=None, range_end: int=None, chunk_size: int=524288, retries=3):
     """
+    General function to get an object from an S3 bucket. One of s3, connection_config, or public_url must be used.
 
+    Parameters
+    ----------
+    url: http str
+        The http url to the file.
+    range_start: int
+        The byte range start for the file.
+    range_end: int
+        The byte range end for the file.
+    chunk_size: int
+        The amount of bytes to download as once.
+
+    Returns
+    -------
+    file object
+        file object of the S3 object.
+    """
+    transport_params = {'buffer_size': chunk_size, 'timeout': 120}
+
+    # Range
+    if (range_start is not None) or (range_end is not None):
+        range_dict = {}
+        if range_start is not None:
+            range_dict['start'] = str(range_start)
+        else:
+            range_dict['start'] = ''
+    
+        if range_end is not None:
+            range_dict['end'] = str(range_end)
+        else:
+            range_dict['end'] = ''
+
+        range1 = 'bytes={start}-{end}'.format(**range_dict)
+        transport_params['headers'] = {'Range': range1}
+    else:
+        range1 = None
+
+    ## Get the object
+    counter = retries
+    while True:
+        try:
+            file_obj = smart_open.open(url, 'rb', transport_params=transport_params, compression='disable')
+            break
+        except Exception as err:
+            counter = counter - 1
+            if counter > 0:
+                sleep(3)
+            else:
+                print('smart_open could not open url with the following error:')
+                print(err)
+                file_obj = None
+                break
+
+    return file_obj
+
+
+def stream_to_file(file_obj: io.BufferedIOBase, file_path: AnyPath, chunk_size=524288):
+    """
+    Convert a file object (stream) to a file on disk. no decompression will occur.
+
+    Parameters
+    ----------
+    file_obj : io.BytesIO or io.BufferedIOBase
+        The file object to be saved to file.
+    file_path: str or pathlib.Path
+        The output file path. If you want the function to do decompression, make sure to include the appropriate compression extension.
+    chunk_size: int
+        The amount of bytes to use in memory for processing the object.
+
+    Returns
+    -------
+    str path
     """
     file_path1 = pathlib.Path(file_path)
     file_path1.parent.mkdir(parents=True, exist_ok=True)
@@ -165,8 +266,22 @@ def stream_to_file(file_obj, file_path, chunk_size=524288):
             chunk = file_obj.read(chunk_size)
 
 
-def decompress_stream_to_file(file_obj, file_path, chunk_size=524288):
+def decompress_stream_to_file(file_obj: io.BufferedIOBase, file_path: AnyPath, chunk_size=524288):
     """
+    Decompress a file object (stream) to a file on disk. Decompression will occur if the file_path has an extension of .zst or .gz, otherwise no decompression will occur. If decompression occurs, then the extension will be removed from the file_path.
+
+    Parameters
+    ----------
+    file_obj : io.BytesIO or io.BufferedIOBase
+        The file object to be saved to file.
+    file_path: str or pathlib.Path
+        The output file path. If you want the function to do decompression, make sure to include the appropriate compression extension.
+    chunk_size: int
+        The amount of bytes to use in memory for processing the object.
+
+    Returns
+    -------
+    str path
 
     """
     file_path1 = pathlib.Path(file_path)
@@ -192,18 +307,32 @@ def decompress_stream_to_file(file_obj, file_path, chunk_size=524288):
     return str(file_path2)
 
 
-def decompress_stream_to_object(file_obj, compression, chunk_size=524288):
+def decompress_stream_to_object(file_obj: io.BufferedIOBase, compression=None, chunk_size=524288):
     """
+    Decompress a file object (stream) to another file object. Decompression will occur only for compression options of zstd or gzip, otherwise no decompression will occur.
+
+    Parameters
+    ----------
+    file_obj : io.BytesIO or io.BufferedIOBase
+        The file object to be saved to file.
+    compression: str
+        The compression types. Options are zstd and gzip.
+    chunk_size: int
+        The amount of bytes to use in memory for processing the object.
+
+    Returns
+    -------
+    file object
 
     """
     b1 = io.BytesIO()
 
-    if compression == 'zstd':
+    if compression.lower() == 'zstd':
         dctx = zstd.ZstdDecompressor()
 
         dctx.copy_stream(file_obj, b1, read_size=chunk_size, write_size=chunk_size)
 
-    elif compression == '.gz':
+    elif compression.lower() == 'gzip':
 
         with gzip.open(file_obj, 'rb') as s_file:
             shutil.copyfileobj(s_file, b1, chunk_size)
@@ -220,7 +349,7 @@ def decompress_stream_to_object(file_obj, compression, chunk_size=524288):
     return b1
 
 
-def put_object_s3(s3, bucket, obj_key, file_obj, metadata=None, content_type=None, retries=3):
+def put_object_s3(s3: botocore.client.BaseClient, bucket: str, obj_key: str, file_obj: io.BufferedIOBase, metadata: dict=None, content_type: str=None, chunk_size=524288, retries=3):
     """
     Function to upload data to an S3 bucket.
 
@@ -232,7 +361,7 @@ def put_object_s3(s3, bucket, obj_key, file_obj, metadata=None, content_type=Non
         The S3 bucket.
     obj_key : str
         The key name for the uploaded object.
-    file_obj : io.BytesIO
+    file_obj : io.BytesIO or io.BufferedIOBase
         The file object to be uploaded.
     metadata : dict or None
         A dict of the metadata that should be saved along with the object.
@@ -264,10 +393,13 @@ def put_object_s3(s3, bucket, obj_key, file_obj, metadata=None, content_type=Non
         transport_params['multipart_upload'] = False
 
     counter = retries
-    while counter > 0:
+    while True:
         try:
             with smart_open.open(s3_url, 'wb', transport_params=transport_params, compression='disable') as f:
-                f.write(file_obj)
+                chunk = file_obj.read(chunk_size)
+                while chunk:
+                    f.write(chunk)
+                    chunk = file_obj.read(chunk_size)
 
             # obj2 = s3.put_object(Bucket=bucket, Key=key, Body=obj, Metadata=metadata, ContentType=content_type)
             break
@@ -287,7 +419,7 @@ def put_object_s3(s3, bucket, obj_key, file_obj, metadata=None, content_type=Non
         #     sleep(3)
 
 
-def put_file_s3(s3, bucket, obj_key, file_path, metadata=None, content_type=None, retries=3):
+def put_file_s3(s3: botocore.client.BaseClient, bucket: str, obj_key: str, file_path: AnyPath, metadata: dict=None, content_type: str=None, retries=3):
     """
     Function to upload data to an S3 bucket.
 
@@ -315,15 +447,32 @@ def put_file_s3(s3, bucket, obj_key, file_path, metadata=None, content_type=None
 
 
 
-def copy_object_s3(s3, source_bucket, dest_bucket, source_key, dest_key, retries=3):
+def copy_object_s3(s3: botocore.client.BaseClient, source_bucket: str, dest_bucket: str, source_key: str, dest_key: str, retries=3):
     """
     Copies an object in an S3 database to another location in an S3 database. They must have the same fundemental connection_config. All metadata is copied to the new object.
     https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.copy_object
+
+    Parameters
+    ----------
+    s3 : boto3.client
+        A boto3 client object
+    source_bucket : str
+        The S3 source bucket.
+    dest_bucket : str
+        The S3 destination bucket.
+    source_key : str
+        The key name for the source object.
+    dest_key : str
+        The key name for the destination object.
+
+    Returns
+    -------
+    None
     """
     source_dict = {'Bucket': source_bucket, 'Key': source_key}
 
     counter = retries
-    while counter > 0:
+    while True:
         try:
             resp = s3.copy_object(Bucket=dest_bucket, Key=dest_key, CopySource=source_dict, MetadataDirective='COPY')
             break
@@ -345,9 +494,22 @@ def copy_object_s3(s3, source_bucket, dest_bucket, source_key, dest_key, retries
     return resp
 
 
-def multi_copy_object_s3(s3, source_bucket, dest_bucket, source_dest_keys, retries=5, threads=30):
+def multi_copy_object_s3(s3: botocore.client.BaseClient, source_bucket: str, dest_bucket: str, source_dest_keys: list, retries=5, threads=30):
     """
     Same as the copy_object_s3 except with multi threading. The input source_dest_keys must be a list of dictionaries with keys named source_key and dest_key.
+
+    Parameters
+    ----------
+    s3 : boto3.client
+        A boto3 client object
+    source_bucket : str
+        The S3 source bucket.
+    source_dest_bucket : list of dict
+        A list of dictionaries with keys named source_key and dest_key. Similar to the copy_obect_s3 function.
+
+    Returns
+    -------
+    None
     """
     keys = copy.deepcopy(source_dest_keys)
 
@@ -366,7 +528,7 @@ def multi_copy_object_s3(s3, source_bucket, dest_bucket, source_dest_keys, retri
     return resp_list
 
 
-def list_objects_s3(s3, bucket, prefix, start_after='', delimiter='', continuation_token='', date_format=None):
+def list_objects_s3(s3: botocore.client.BaseClient, bucket: str, prefix: str, start_after: str='', delimiter: str='', continuation_token: str='', date_format: str=None):
     """
     Wrapper S3 function around the list_objects_v2 base function with a Pandas DataFrame output.
 
@@ -432,7 +594,7 @@ def list_objects_s3(s3, bucket, prefix, start_after='', delimiter='', continuati
     return f_df1
 
 
-def list_object_versions_s3(s3, bucket, prefix, start_after='', delimiter=None, date_format=None):
+def list_object_versions_s3(s3: botocore.client.BaseClient, bucket: str, prefix: str, start_after: str='', delimiter: str=None, date_format: str=None):
     """
     Wrapper S3 function around the list_object_versions base function with a Pandas DataFrame output.
 
